{-# LANGUAGE ViewPatterns      #-}
{-# LANGUAGE OverloadedStrings #-}
{-# LANGUAGE LambdaCase #-}
{-# LANGUAGE ScopedTypeVariables #-}
{-# LANGUAGE TypeApplications #-}
{-# LANGUAGE BangPatterns #-}

module Gargantext.API.Server.Named.Remote (
  remoteAPI
  ) where

import Codec.Serialise
import Conduit
import Control.Exception.Safe qualified as Safe
import Control.Exception (toException)
import Control.Lens (view, (#))
import Control.Monad.Except (throwError, MonadError)
import Data.Aeson qualified as JSON
import Data.ByteString.Builder qualified as B
import Data.ByteString.Lazy qualified as BL
import Data.Conduit.Combinators qualified as C
import Data.Conduit.List qualified as CL
import Data.Foldable (foldlM)
import Data.Text qualified as T
import Gargantext.API.Admin.Auth
import Gargantext.API.Admin.Auth.Types (AuthenticatedUser(..))
import Gargantext.API.Auth.PolicyCheck (remoteExportChecks)
import Gargantext.API.Errors.Types
import Gargantext.API.Ngrams.Prelude (getNgramsList)
import Gargantext.API.Ngrams.Types (NgramsList)
import Gargantext.API.Node.Document.Export.Types (DocumentExport)
import Gargantext.API.Prelude (IsGargServer)
import Gargantext.API.Routes.Client (remoteImportClient)
import Gargantext.API.Routes.Named.Remote qualified as Named
import Gargantext.Core.Config
import Gargantext.Core (lookupDBid)
import Gargantext.Core.Types.Main
import Gargantext.Database.Admin.Types.Node
import Gargantext.Database.Prelude (IsDBCmd)
import Gargantext.Database.Query.Table.Node.Error (HasNodeError, nodeError, NodeError (..))
import Gargantext.Database.Query.Table.Node (insertNodeWithHyperdata, getNodes, getUserRootPrivateNode)
import Gargantext.Database.Schema.Node (NodePoly(..))
import Gargantext.Orphans ()
import GHC.Generics (Generic)
import Prelude
import Servant.Client.Streaming (mkClientEnv, withClientM, ClientError)
import Servant.Server.Generic (AsServerT)
import Gargantext.API.Node.Document.Export (get_document_json)

data ExportableNode =
  ExportableNode {
      _en_node :: Node JSON.Value
    -- | If this node is a \"docs\" node, remotely export also
    -- all the associated documents.
    , _en_associated_docs :: Maybe DocumentExport
    -- | If this node is a \"terms\" node, remotely export also
    -- all the associated ngrams
    , _en_associated_ngrams :: Maybe NgramsList
    } deriving Generic

instance Serialise ExportableNode where

remoteAPI :: (MonadIO m, IsGargServer env BackendInternalError m)
          => AuthenticatedUser
          -> Named.RemoteAPI (AsServerT m)
remoteAPI authenticatedUser = Named.RemoteAPI $
  Named.RemoteAPI'
    { remoteExportEp = \payload@Named.RemoteExportRequest{..} mgr ->
        withPolicy authenticatedUser (remoteExportChecks _rer_node_id) (remoteExportHandler authenticatedUser payload) mgr
    , remoteImportEp = remoteImportHandler authenticatedUser
    }

type ExpectedPayload = Tree ExportableNode

remoteImportHandler :: forall err env m.
                    (HasNodeError err, HasBackendInternalError err, IsDBCmd env err m, MonadIO m)
                    => AuthenticatedUser
                    -> ConduitT () Named.RemoteBinaryData IO ()
                    -> m [NodeId]
remoteImportHandler loggedInUser c = do
  chunks <- liftIO $ sourceToList $ c .| C.map (B.byteString . Named.getRemoteBinaryData)
  -- FIXME(adn): We have to find a way to deserialise this into a streaming fashion and
  -- attempt insertion one element of the list at the time.
  case deserialiseOrFail @ExpectedPayload (B.toLazyByteString $ mconcat chunks) of
    Left err -> throwError $ _BackendInternalError # InternalUnexpectedError (toException $ userError $ "Deserialization error: " ++ show err)
    Right (TreeN x xs) -> do
      -- NOTE(adn) By default, we append the imported node(s) to the user's
      -- private folder.
      privateFolderId <- _node_id <$> getUserRootPrivateNode (_auth_user_id loggedInUser)
      -- Attempts to insert nodes a we go along.
      rootNode <- insertNode (Just privateFolderId) x
      foldlM (insertTrees (Just rootNode)) [rootNode] xs
  where

    insertNode :: Maybe NodeId -> ExportableNode -> m NodeId
    insertNode mb_parent (ExportableNode x _mb_docs _mb_terms) = case lookupDBid $ _node_typename x of
      Nothing -> throwError $ _BackendInternalError # InternalUnexpectedError (toException $ userError $ "remoteImportHandler: impossible, node with invalid type.")
      Just ty -> do
        new_node <- insertNodeWithHyperdata ty (_node_name x) (_node_hyperdata x) mb_parent (_auth_user_id loggedInUser)
        pure new_node

    insertTrees :: Maybe NodeId -> [NodeId] -> Tree ExportableNode -> m [NodeId]
    insertTrees currentParent !acc (TreeN x xs) = do
      childrenRoot <- insertNode currentParent x
      (`mappend` acc) <$> foldlM (insertTrees (Just childrenRoot)) [childrenRoot] xs



remoteExportHandler :: ( MonadIO m, Safe.MonadCatch m
                       , IsGargServer err env m
                       )
                    => AuthenticatedUser
                    -> Named.RemoteExportRequest
                    -> m [NodeId]
remoteExportHandler loggedInUser Named.RemoteExportRequest{..} = do
  mgr  <- view gargHttpManager
  nodes <- getNodes _rer_node_id
  checkNodesTypeAllowed nodes
  exportable <- makeExportable (_auth_node_id loggedInUser) nodes
  liftIO (withClientM (remoteImportClient _rer_instance_auth (streamEncoder exportable)) (mkClientEnv mgr _rer_instance_url) streamDecode)
              `Safe.catch` \(e :: BackendInternalError) -> throwError $ _BackendInternalError # e

makeExportable :: IsGargServer err env m
               => NodeId
               -> Tree (Node JSON.Value)
               -> m (Tree ExportableNode)
makeExportable userNodeId (TreeN x xs)
  | Just nty <- lookupDBid (_node_typename x)
  = do
    mb_docs <- case nty of
      NodeDocument -> Just <$> get_document_json userNodeId (_node_id x)
      _            -> pure Nothing
    mb_ngrams <- case nty of
      NodeList -> Just <$> getNgramsList (_node_id x)
      _        -> pure Nothing
    let exportableRoot = ExportableNode x mb_docs mb_ngrams
    children <- mapM (makeExportable userNodeId) xs
    pure $ TreeN exportableRoot children
  | otherwise
  = throwError $ _BackendInternalError
               # InternalUnexpectedError (toException $ userError $ "remoteImportHandler: impossible, node with invalid type.")

checkNodesTypeAllowed :: (MonadError e m, HasNodeError e) => Tree (Node a) -> m ()
checkNodesTypeAllowed (TreeN r xs) = do
  checkNodeTypeAllowed r
  mapM_ checkNodesTypeAllowed xs

checkNodeTypeAllowed :: (MonadError e m, HasNodeError e) => Node a -> m ()
checkNodeTypeAllowed n
  | Just nty <- lookupDBid (_node_typename n)
  , nty `elem` exportableNodeTypes
  = pure ()
  | otherwise
  = let msg = "It's possible to export only the following node of type: " <> T.intercalate "," (map (T.pack . show) exportableNodeTypes)
    in nodeError $ NodeNotExportable (_node_id n) msg

-- | At the moment we support only export corpus nodes and their children (i.e. "Docs", "Terms", "Graph").
exportableNodeTypes :: [NodeType]
exportableNodeTypes = [ NodeCorpus, NodeCorpusV3, NodeTexts, NodeGraph, NodeList, NodePhylo ]

streamEncoder :: MonadIO m => ExpectedPayload -> ConduitT () Named.RemoteBinaryData m ()
streamEncoder = CL.sourceList . map Named.RemoteBinaryData . BL.toChunks . serialise

-- | Returns a conduit which can be used to decode
streamDecode :: Either ClientError [NodeId] -> IO [NodeId]
streamDecode = \case
  Left err -> Safe.throwIO $ InternalUnexpectedError (toException $ userError $ show err)
  Right x  -> pure x
